@@ -1,10 +1,10 @@
 ;;; outline-indent.el --- Folding text based on indentation (origami alternative) -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2024-2025 James Cherti | https://www.jamescherti.com/contact/
+;; Copyright (C) 2024-2026 James Cherti | https://www.jamescherti.com/contact/
 
 ;; Author: James Cherti
-;; Package-Version: 1.1.5
-;; Package-Revision: 1.1.5-0-g46cb8657bed0
+;; Package-Version: 1.1.7
+;; Package-Revision: 1.1.7-0-g75e1b5d41e7b
 ;; URL: https://github.com/jamescherti/outline-indent.el
 ;; Keywords: outlines
 ;; Package-Requires: ((emacs "26.1"))
@@ -90,18 +90,24 @@
   "Default indentation offset.
 If nil, the offset is automatically determined based on the current mode.
 This value is used to calculate outline levels from the current indentation."
-  :type '(choice (integer :tag "Custom width")
-                 (const :tag "Auto detect" nil))
+  :type '(choice (natnum :tag "Custom width")
+                 (natnum :tag "Auto detect" nil))
   :group 'outline-indent)
 
 (defcustom outline-indent-shift-width nil
   "Shift width used for indentation adjustments during promotion and demotion.
 If nil, this value defaults to `outline-indent-default-offset', which is
 automatically determined according to the current mode."
-  :type '(choice (integer :tag "Custom width")
+  :type '(choice (natnum :tag "Custom width")
                  (const :tag "Auto detect" nil))
   :group 'outline-indent)
 
+(defcustom outline-indent-maximum-level nil
+  "Maximum allowed depth for outline indentation.
+If nil, there is no maximum."
+  :type '(choice (natnum :tag "Maximum level")
+                 (const :tag "No limit" nil)))
+
 (defcustom outline-indent-ellipsis nil
   "String used as the ellipsis character in `outline-indent-mode'.
 
@@ -364,13 +370,41 @@ follow the mode-specific coding style automatically."
       (unless outline-indent-shift-width
         (setq-local outline-indent-shift-width major-mode-offset)))))
 
+(defun outline-indent-level-ignore-empty-lines ()
+  "Determine the outline level based on the current indentation."
+  (let* ((indentation-string (match-string 1))
+         (indentation-width (if indentation-string
+                                (string-width indentation-string)
+                              0))
+         (depth (if (and (= indentation-width 0)
+                         (save-excursion
+                           (beginning-of-line)
+                           (looking-at-p "^\\s-*$")))
+                    0
+                  (+ 1 (/ indentation-width
+                          (max (if outline-indent-default-offset
+                                   outline-indent-default-offset
+                                 1)
+                               1))))))
+    (if outline-indent-maximum-level
+        (min depth (1+ outline-indent-maximum-level))
+      depth)))
+
 (defun outline-indent-level ()
   "Determine the outline level based on the current indentation."
-  (+ 1 (/ (current-indentation)
-          (max (if outline-indent-default-offset
-                   outline-indent-default-offset
-                 1)
-               1))))
+  (let* ((indentation-width (current-indentation))
+         ;; TODO fix this
+         ;; (indentation-string (match-string 1))
+         ;; (indentation-width (if indentation-string
+         ;;                        (string-width indentation-string)
+         ;;                      0))
+         (depth (1+ (/ indentation-width
+                       (max (or outline-indent-default-offset
+                                1)
+                            1)))))
+    (if outline-indent-maximum-level
+        (min depth (1+ outline-indent-maximum-level))
+      depth)))
 
 (defun outline-indent--update-ellipsis ()
   "Update the buffer's outline ellipsis."
@@ -441,7 +475,7 @@ to insert content at the same indentation level after the current fold."
         (forward-line -1))
       (indent-to initial-indentation)))
 
-  (when (and (bound-and-true-p evil-mode)
+  (when (and (bound-and-true-p evil-local-mode)
              (fboundp 'evil-insert-state))
     (funcall 'evil-insert-state)))
 
@@ -456,7 +490,10 @@ addressing the issue where the cursor might be reset after the operation."
   (outline-indent-move-subtree-down (- arg)))
 
 (defun outline-indent--deactivate-region ()
-  "Deactivate the current region and move point to the start of the region."
+  "Deactivate the current region and relocate point to the start.
+If a region is currently active, moves the point to the beginning of the region
+and disables the mark. This ensures a consistent starting state for subsequent
+selection operations."
   (when (use-region-p)
     (goto-char (region-beginning))
     (deactivate-mark)))
@@ -722,24 +759,41 @@ Stop at the first and last indented blocks of a superior indentation."
 ;;;###autoload
 (defun outline-indent-select ()
   "Select the indented block at point.
-Selects the entire indented block at point, activating a visual region that
-spans the heading and all of its associated indented content."
+Identifies the heading and all associated indented content, then activates a
+visual region spanning from the heading start to the end of the block."
   (interactive)
   (outline-indent--deactivate-region)
   (let ((begin (save-excursion
                  (outline-back-to-heading)
                  (point)))
         (end (outline-indent--next-lower-or-equal-indentation)))
-    (goto-char (+ end 1))
-    (push-mark)
+    (push-mark (1+ end) nil t)
     (goto-char begin)
     (activate-mark)))
 
 ;;;###autoload
 (defun outline-indent-close-folds ()
-  "Close all folds."
+  "Close all folds and ensure the first heading remains visible."
   (interactive)
-  (with-no-warnings (outline-hide-sublevels 1)))
+  (with-no-warnings
+    (outline-hide-sublevels 1))
+
+  (let ((heading-point (save-excursion
+                         (condition-case nil
+                             (progn
+                               (goto-char (window-start))
+                               (beginning-of-visual-line)
+                               (if (outline-on-heading-p)
+                                   (point)
+                                 (outline-back-to-heading t)))
+                           (error
+                            nil)))))
+    ;; Ensure folded headings remain visible after hiding subtrees. Fixes a
+    ;; bug in outline and Evil where headings could scroll out of view when
+    ;; their subtrees were folded. TODO Send a patch to Emacs and/or Evil
+    (when (and heading-point
+               (< heading-point (window-start)))
+      (set-window-start (selected-window) heading-point t))))
 
 ;;;###autoload
 (defun outline-indent-open-folds ()
@@ -751,42 +805,89 @@ spans the heading and all of its associated indented content."
 (defun outline-indent-open-fold ()
   "Open fold at point."
   (interactive)
-  (condition-case nil
-      (let ((on-visible-heading (when (outline-on-heading-p t)
-                                  (outline-invisible-p))))
-        (save-excursion
-          (while (outline-indent-folded-p)
-            ;; Repeatedly reveal children and body until the entry is no
-            ;; longer folded
-            (save-excursion
-              (outline-back-to-heading)
-              (outline-show-children)
-              (outline-indent--legacy-outline-show-entry))))
+  ;; Workaround for an outline-mode limitation: when jumping via imenu or
+  ;; search, sibling headings above the current one and at the same level
+  ;; often remain hidden. This ensures all sub-items at the current level are
+  ;; revealed, preventing the 'isolated item' effect.
+  (save-excursion
+    ;; Climbing as long as a parent heading exists
+    (catch 'done
+      (while (not (bobp))
+        (condition-case nil
+            (progn (outline-up-heading 1 t))
+          (error
+           (throw 'done t)))
 
-        (when on-visible-heading
-          (outline-indent--legacy-outline-hide-subtree)))
+        (outline-show-children))))
+
+  ;; Repeatedly reveal children and body until the entry is no longer folded
+  (condition-case nil
+      (progn
+        (let* ((on-invisible-heading (when (outline-on-heading-p t)
+                                       (outline-invisible-p)))
+               (on-visible-heading (save-excursion
+                                     (beginning-of-line)
+                                     (outline-on-heading-p))))
+          (when (use-region-p)
+            (outline-indent-select))
+
+          (save-excursion
+            (while (outline-indent-folded-p)
+              ;; Repeatedly reveal children and body until the entry is no
+              ;; longer folded
+              (save-excursion
+                (outline-back-to-heading)
+                (outline-show-children)
+                (outline-indent--legacy-outline-show-entry))))
+
+          ;; If the header was previously hidden, hide the subtree to collapse
+          ;; it. Otherwise, leave the fold open. This allows the user to decide
+          ;; whether to expand the content under the cursor.
+          (when (and on-invisible-heading (not on-visible-heading))
+            (outline-indent--legacy-outline-hide-subtree))))
     ;; Ignore `outline-before-first-heading'
     (outline-before-first-heading
      nil)))
 
 ;;;###autoload
 (defun outline-indent-close-fold ()
-  "Close fold at point."
+  "Close the current heading's subtree in a robust manner.
+
+If the current heading is folded or contains no content, move to the previous
+heading with a higher level and close its subtree.
+
+Otherwise, close the current subtree. Ensures that folded headings remain
+visible in the window after hiding."
   (interactive)
   (condition-case nil
       (save-excursion
+        ;; Move to the current heading; error if before the first heading
         (outline-back-to-heading)
-        (if (or (outline-indent-folded-p)  ; Folded?
-                ;; Fold without any content
-                (let ((start (save-excursion (end-of-line) (point)))
-                      (end (save-excursion (outline-end-of-subtree) (point))))
-                  (= start end)))
-            (progn
-              (when (eq (ignore-errors (outline-up-heading 1 t) :success)
-                        :success)
-                (when (outline-on-heading-p)
-                  (outline-indent--legacy-outline-hide-subtree))))
-          (outline-indent--legacy-outline-hide-subtree)))
+
+        (let ((heading-point (point)))
+          ;; If the current heading is folded, or if it contains no content,
+          ;; move to the previous higher-level heading.
+          (when (or (outline-indent-folded-p)  ; Folded?
+                    ;; Fold without any content
+                    (let ((start (save-excursion (end-of-line)
+                                                 (point)))
+                          (end (save-excursion (outline-end-of-subtree)
+                                               (point))))
+                      (= start end)))
+            ;; Try to move up to previous higher-level heading
+            (outline-up-heading 1 t)
+            (setq heading-point (point)))
+
+          (when (outline-on-heading-p)
+            (outline-indent--legacy-outline-hide-subtree))
+
+          ;; Ensure folded headings remain visible after hiding subtrees. Fixes
+          ;; a bug in outline and Evil where headings could scroll out of view
+          ;; when their subtrees were folded.
+          ;; TODO Send a patch to Emacs and/or Evil
+          (when (and heading-point
+                     (< heading-point (window-start)))
+            (set-window-start (selected-window) heading-point t))))
     ;; Ignore `outline-before-first-heading'
     (outline-before-first-heading
      nil)))
@@ -882,10 +983,16 @@ WHICH is ignored (backward compatibility with `outline-promote')."
 
 ;;; Mode
 
+;; (defun outline-indent--heading-regexp ()
+;;   "Compute heading regexp based on leading indentation."
+;;   (rx-to-string
+;;    `(and line-start
+;;          (group (zero-or-more (any " \t")))
+;;          (not (any " \t\n")))))
+
 ;;;###autoload
 (define-minor-mode outline-indent-minor-mode
-  "Toggle `outline-indent-minor-mode'.
-This mode sets up outline to work based on indentation."
+  "Minor mode for folding text according to indentation levels."
   :lighter " OutlInd"
   :keymap outline-indent-minor-mode-map
   :group 'outline-indent
@@ -909,8 +1016,9 @@ This mode sets up outline to work based on indentation."
         (setq-local outline-heading-alist nil)
         (setq-local outline-level #'outline-indent-level)
         (setq-local outline-heading-end-regexp "\n")
-        (setq-local outline-regexp (rx bol
-                                       (zero-or-more (any " \t"))
+        ;; (setq-local outline-regexp (outline-indent--heading-regexp))
+        (setq-local outline-regexp (rx line-start
+                                       (group (zero-or-more (any " \t")))
                                        (not (any " \t\n"))))
         (outline-indent--update-ellipsis)
         (outline-indent--setup-basic-offset)
